#ifndef LibAard_h
#define LibAard_h

#include <queue>
#include <termios.h>
#include <unistd.h>

class AardRemoteMessage;

// Class for interface with AardRemote on linux.
//
class LibAard {
public:

   LibAard();

   virtual
   ~LibAard();

      // Open the serial device to talk to the remote. Typically this
      // is something like /dev/ttyACM0 for a USB connection or
      // /dev/rfcomm0 for a bluetooth connection.
      //
      // The remote microcode expects 115200 baud, so that is hard
      // coded into the open.
      //
      // Returns the file descriptor or -1 if there was a problem.
      // You can arrange to hook the file descriptor into some event
      // loop should you so desire in order to know when it might be
      // a good time to call PollMessage().
      //
   int
   OpenAardRemote(const char * devname);

      // Close the connection to the remote.
      //
   void
   CloseAardRemote();

      // Opening the device may or may not reset the remote. If it does
      // reset the remote, it will be a while before it is up and taking
      // again, so this routine is designed to be called after the open to
      // wait till the remote is willing to speak.
      //
      // It operates by sending a bad command byte and waiting for an error
      // response or an async MKAS_RESET message to come in. If the device
      // was reset by the open, the reset message will eventually show up, if
      // the device was not reset by the open, the error response from the
      // bad command byte will show up.
      //
      // The return code is the message code for which message it got.
      //
      // Note: I get the impression that sometimes the device resets itself
      // twice in a row, so you might want to call this in a loop until
      // you see two or more consecutive MKAS_BILGE responses.
      //
   int
   WaitForStartup();

      // Send just a command with no associated data.
      //
   void
   SendCommand(unsigned char cmdcode);

      // Send a command with a 16 bit chunk of data.
      //
   void
   SendCommand(unsigned char cmdcode, unsigned short cmdata);

      // Send an arbitrary length command.
      //
   void
   SendCommand(unsigned char cmdcode, const unsigned short * cmdata, int len);

      // Send an arbitrary length command.
      //
   void
   SendCommand(unsigned char cmdcode, const std::vector<unsigned short> & dat);

      // Check for data coming in on the file descriptor. If enough has shown
      // up to form a complete message, return a message object (which should
      // be deleted by the caller when it is done with it).
      //
      // If no message is available, return NULL.
      //
   AardRemoteMessage *
   PollMessage();

      // Wait for a message to come in for up to the given number of
      // milliseconds. If a message has not arrived in that time, return NULL.
      //
      // If milliseconds is 0, wait forever.
      //
      // The returned message object should be deleted by the caller.
      //
   AardRemoteMessage *
   WaitMessage(unsigned int milliseconds);

private:

   int fd;

   struct termios orig_ts;
   bool orig_ts_valid;

   AardRemoteMessage * incoming;
   bool have_leftover_byte;
   unsigned char leftover_byte;
   std::queue<AardRemoteMessage *> outbox;

   ssize_t
   write_and_log(const void * buf, size_t count);

   ssize_t
   read_and_log(void * buf, size_t count);

   void
   IncomingToOutbox();
};

#endif // LibAard_h
